Set Implicit Arguments.
Require Import MyTactics.
Require Import TyCon.
Require Import DeBruijn.
Require Import Syntax.
Require Import WkInversion.
Require Import Signature.
Require Import Exclusive.

(* ---------------------------------------------------------------------------- *)

(* We define a relation [adopts T U], which means that an object of type [T] can
   adopt an object of type [U]. *)

Inductive adopts: term -> term -> Prop :=

(* An algebraic data type [tc] can adopt an object of type [U], provided the
   [adopts] clause in the definition of [tc] mentions [U], and provided [tc]
   and [U] are exclusive. *)

| AdoptsCon:
    forall tc U,
    tc_adopts tc = U ->
    tc_mode tc = Exclusive ->
    exclusive U ->
    adopts (TyCon tc) U

(* A structural type [TyBlock Exclusive _ _ U] can adopt an object of type [U],
   provided [U] is exclusive. *)

| AdoptsBlock:
    forall tag Ts U,
    exclusive U ->
    adopts (TyBlock Exclusive tag Ts U) U.

Hint Constructors adopts : adopts.

Ltac select_adopts action :=
  match goal with h: adopts _ _ |- _ => action h end.
Ltac adopts_inversion :=
  gji select_adopts.

(* ---------------------------------------------------------------------------- *)

(* If [T] adopts [U] and if [T] is well-formed, then [U] is well-formed. *)

Lemma adopts_wk:
  forall T U,
  adopts T U ->
  forall K,
  wk K T TYPE ->
  wk K U TYPE.
Proof.
  inversion 1; intros; subst; wk_inversion; eauto using tc_adopts_wk, wk_empty.
Qed.

Hint Resolve adopts_wk : wk.

(* ---------------------------------------------------------------------------- *)

(* If [T] adopts [U], then [T] and [U] are exclusive. *)

Lemma adopters_are_exclusive:
  forall T U,
  adopts T U ->
  exclusive T.
Proof.
  induction 1; econstructor; eauto.
Qed.

Lemma adoptees_are_exclusive:
  forall T U,
  adopts T U ->
  exclusive U.
Proof.
  induction 1; eauto.
Qed.

Hint Resolve adopters_are_exclusive adoptees_are_exclusive : exclusive.

