Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.
Require Import Typing.
Require Import Environments.
Require Import Harmless.
Require Import JWeakening.
Require Import JSubstitution.

(* DeBruijn macro for swapping two consecutive indices. *)

Definition swap (n: nat) (t: term): term :=
  subst (var (S n)) n (shift (S (S n)) t).

(* As [rotate1], swap is self inverse, and the very same proof
   gives this more general version. *)

Lemma swap_self_inverse:
  forall t n,
  swap n (swap n t) = t.
Proof.
  intros; unfold swap.
  rewrite lift_subst_2 by omega.
  rewrite <- lift_lift by omega.
  rewrite lift_var. lift_idx.
  rewrite subst_subst by omega.
  rewrite subst_var.
  rewrite lift_var. subst_idx.
  replace (S n - 1) with n; auto with arith.
  replace (TVar (shift n (S n))) with (shift (S n) (TVar (S n))).
  rewrite <- lift_subst_2 by omega.
  rewrite pun_2.
  replace (TVar (S n)) with (var (S n)); auto.
  rewrite pun_2.
  auto.
  replace (TVar (S n)) with (var (S n)); auto.
  rewrite lift_var.
  lift_idx.
  lift_idx.
  auto.
Qed.

(* [swap] interacts with [shift] the following way. *)

Lemma swap_shift:
  forall t n,
  swap n (shift n t) = shift (S n) t.
Proof.
  intros; unfold swap.
  replace (S (S n)) with (1 + (S n)); auto with arith.
  rewrite <- lift_lift; auto.
  rewrite subst_lift.
  auto.
Qed.

Lemma swap_shift_2:
  forall t n,
  swap n (shift (S n) t) = shift n t.
Proof.
  intros.
  rewrite <- swap_shift.
  eauto using swap_self_inverse.
Qed.

(* Relation between [swap] and [rotate1]. *)

Lemma swap_rotate1:
  forall t,
  rotate 1 t = swap 0 t.
Proof.
  auto.
Qed.

(* The [swap_shift] lemmas, translated for [rotate1]. *)

Lemma rotate_shift_1:
  forall t,
  rotate 1 (shift 1 t) = shift 0 t.
Proof.
  intros.
  rewrite swap_rotate1.
  eauto using swap_shift_2.
Qed.

Lemma rotate_shift_0:
  forall t,
  rotate 1 (shift 0 t) = shift 1 t.
Proof.
  intro.
  rewrite <- rotate_shift_1.
  eauto using rotate1_self_inverse.
Qed.

(* The first two variables of a kinding environment can be swapped. *)

Lemma j_rotate1:
  forall n R k1 k2 K P t T P' t' T',
  jt n R (insert 0 k1 (insert 0 k2 K)) P t T ->
  P' = rotate 1 P ->
  t' = rotate 1 t ->
  T' = rotate 1 T ->
  substitutable k1 ->
  jt n R (insert 0 k2 (insert 0 k1 K)) P' t' T'.
Proof.
  introv hjt ? ? ? ?.
  (* The hypothesis is weakened: a [k1] is inserted at position 2. *)
  forwards hjtw: (get4 j_weakening) hjt;
    repeat rewrite <- insert_insert; eauto.
  (* The new [k1] replaces the old one. *)
  forwards hjtws: (get4 j_substitution) hjtw;
    eauto using wk_weakening, lookup_insert_bingo with wk lookup_insert.
  simpl_lift_in hjtws.
  subst; unfold rotate.
  eauto with omega.
Qed.

