Set Implicit Arguments.
Require Import MyTactics.
Require Import TyCon.
Require Import DeBruijn.
Require Import Syntax.
Require Import Signature.

(* ---------------------------------------------------------------------------- *)

(* We define a predicate [exclusive T], which indicates which types are
   exclusive. *)

(* Adopters and adoptees must have exclusive types. *)

Inductive exclusive: term -> Prop :=

(* An algebraic data type [tc] is exclusive if its definition says so. *)

| ExclusiveCon:
    forall tc,
    tc_mode tc = Exclusive ->
    exclusive (TyCon tc)

(* A structural type is exclusive if it says so. *)

| ExclusiveBlock:
    forall tag Ts U,
    exclusive (TyBlock Exclusive tag Ts U).

Hint Constructors exclusive : exclusive.

(* ---------------------------------------------------------------------------- *)

(* Basic properties. *)

Lemma exclusive_shift:
  forall T,
  exclusive T ->
  forall x,
  exclusive (shift x T).
Proof.
  induction 1; intros; simpl_lift_goal; econstructor; eauto.
Qed.

Hint Resolve exclusive_shift : j_weakening.

Lemma exclusive_subst:
  forall T,
  exclusive T ->
  forall w x,
  exclusive (subst w x T).
Proof.
  induction 1; intros; simpl_subst_goal; econstructor; eauto.
Qed.

Hint Resolve exclusive_subst : j_substitution.

