Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import DBSwap.
Require Import Syntax.
Require Import Environments.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import Top.
Require Import JBasic.
Require Import JCanonicalization.
Require Import JDecomposition.
Require Import WkInversion.
Require Import Dup.
Require Import Intro.

(* ---------------------------------------------------------------------------- *)

(* A universal quantifier can be distributed over the type of a harmless
   function. *)

Lemma harmless_distrib:
  forall R K t kappa T U,
  harmless t ->
  consistent R ->
  jctop R K (TAbs t) (TyForall kappa (TyArrow T U)) ->
  jctop R K (TAbs t) (TyArrow (TyForall kappa T) (TyForall kappa U)).
Proof.
  introv hl hc hj.
  jpi_inversion.
  revert dependent R2.
  dependent induction H2; intros.

  (* Frame. *)
  rewrite reconfigure00.
  eapply IHj; auto.
    eapply consistent_by_equality; eauto with aac.
    rewrite star_commutative; eauto with j.

  (* The real case. *)
  clear IHj.
  (* Forwards: use the typing hypothesis on [TAbs t] to build a typing
     judgment for [t]. *)
  jpi_inversion.
  make_jt_hypothesis hj.
  forwards hjc: canonicalization_preliminary hj; eauto with j_weakening.
  simpl_lift_in hjc.
  decomposition.
  (* Reduce the goal to a typing for [t]. *)
  econstructor; eauto using @core_right_unit, JpiEmpty with wk.
  eapply JFunction.
    assert (wk (insert 0 kappa K) (TyArrow T U) TYPE).
    eauto with wk.
    wk_inversion.
    eauto with wk.
    eauto using DupEmpty.
    unfold intro.
    simpl_lift_goal.
    eapply JSubLeft; eauto using JpsDisappearsLeft.
    eapply JSubLeft; eauto using JpsForallType2Perm.
    eapply JForallIntro.
    wk_inversion; auto.
    auto.
    simpl_lift_goal.
    rewrite plus_0_r in *.
  (* The connection between the typing hypothesis for [t] and the goal comes
     from the swapping lemma. *)
    eapply j_rotate1; eauto using rotate_shift_1, rotate_shift_0 with substitutable.
    rewrite <- rotate1_self_inverse at 1.
    unfold rotate at 2.
    simpl_lift_goal; simpl_subst_goal.
    rewrite <- rotate_shift_1.
    unfold rotate. auto with arith.

Grab Existential Variables.
auto. auto.

Qed.
