Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.

(* ---------------------------------------------------------------------------- *)

(* A notation for introducing a new variable, together with a permission for
   it. Informally, [intro Q T] represents the permission [Q * (x @ T)], where
   [x] is ``a new variable''. *)

Definition intro Q T :=
  TyStar (shift 0 Q) (TyAnchored (TVar 0) (shift 0 T)).

(* ---------------------------------------------------------------------------- *)

(* [intro] preserves well-kindedness. *)

Lemma intro_wk:
  forall K Q T,
  wk K Q PERM ->
  wk K T TYPE ->
  wk (insert 0 VALUE K) (intro Q T) PERM.
Proof.
  unfold intro. eauto 7 using wk_weakening with wk.
Qed.

Hint Resolve intro_wk : wk.

(* [intro] commutes with [shift]. *)

Lemma shift_intro:
  forall sx Q T x,
  sx = 1 + x ->
  shift sx (intro Q T) = intro (shift x Q) (shift x T).
Proof.
  unfold intro. intros; subst; simpl_lift_goal; eauto with f_equal lift_lift.
Qed.

Hint Resolve shift_intro : j_weakening.

(* [intro] commutes with [subst]. *)

Lemma subst_intro:
  forall Q T U x U' sx,
  shift 0 U = U' ->
  1 + x = sx -> 
  subst U' sx (intro Q T) = intro (subst U x Q) (subst U x T).
Proof.
  unfold intro. intros; subst. simpl_subst_goal. eauto with f_equal lift_subst.
Qed.

Hint Resolve subst_intro : j_substitution.

