Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Row.
Require Import WkInversion.
Require Import Intro.
Require Import Dup.
Require Import Typing.
Require Import JBasic.
Require Import Top.
Require Import JWeakening.
Require Import JSubstitution.
Require Import SubsumptionSoundness.

(* ---------------------------------------------------------------------------- *)

(* A canonicalization lemma. If the value [v] is well-typed according to
   [jt], then it must be well-typed according to [jc]. In other words,
   an unrestricted type derivation can be turned into a canonical type
   derivation. The permission subsumption steps are eliminated by appeal to
   the lemma [jps_sound]. *)

(* The proof is by well-founded induction over the height of the type
   derivation -- as opposed to by structural induction over the type
   derivation -- because [JExistsElim] requires that the substitution lemma be
   applied to the premise before the induction hypothesis for the premise is
   invoked. *)

Lemma canonicalization_preliminary:
  forall n R1 K P1 v T,
  jt n R1 K P1 v T ->
  forall R2 R,
  jpi R2 K P1 ->
  star R1 R2 = R ->
  consistent R ->
  wk K v VALUE ->
  jctop R K v T.
Proof.
  intro n. induction_wf ih: lt_wf n.
  introv h; dependent destruction h; intros; subst;
  wk_inversion; jpi_inversion;
  try solve [ eauto using JcTopSingleton, JFunction, (JBlock 0) with wk jpi ].
  (* JFrame *)
  (* Apply the induction hypothesis; conclude by using [JcTopFrame]. *)
  solve [ eauto using JcTopFrame, consistent_1p23_12 with aac ].
  (* JCut *)
  (* Apply the induction hypothesis, and we are done. *)
  solve [ eauto with j aac ].
  (* JSubLeft *)
  (* Apply [jps_sound], then apply the induction hypothesis, and we are done. *)
  solve [ eauto using jps_sound with consistent_split ].
  (* JSubRight *)
  (* Apply the induction hypothesis, then apply [sub_sound], and we are done. *)
  solve [ eauto using sub_sound ].
  (* JFold *)
  solve [ eauto using JcTopFold ].
  (* JDynamic *)
  eapply JcTopDynamic.
  solve [ eauto using @SCAM_limit_split_left with lt_lt typeclass_instances ].
  (* JLck *)
  forwards: lck_resource_maps_star; eauto. unpack.
  solve [ eauto using JcTopLck ].
    (* Lock are no longer maximal. *)
    (* solve [ eauto using JcTopLck with SCAM_maps_maximal_lck ]. *)
  (* JLocked *)
  solve [ eauto using JcTopLocked with SCAM_maps_maximal_lck ].
(*  (* JExistsElim *)
  (* The term [t] is well-typed in the presence of an abstract name [x] and
     under a precondition [P] which may refer to [x]. Now, we know that [P]
     holds when [x] is replaced with [v]. Thus, we replace [x] with [v]
     in the type derivation for [t]. The term [t] and its type [T] actually
     do not refer to [x], so they are unaffected. *)
  substitution.
  (* Applying the induction hypothesis to this new type derivation for [t]
     yields the goal. *)
  solve [ eauto ].*)
  (* JForallIntro *)
  (* We are establishing canonicalization under an arbitrary environment [K],
     so the induction hypothesis is strong enough to take us under this
     capital Lambda. *)
  eauto 6 using JcTopForallIntro, (last6 j_weakening), wk_weakening.
Qed.

(* The following corollary is more readable. It is unused. *)

Lemma canonicalization:
  forall n R v T,
  jt n R zilch TyEmpty v T ->
  consistent R ->
  wk zilch v VALUE ->
  jctop R zilch v T.
Proof.
  eauto using canonicalization_preliminary, @core_right_unit with j.
Qed.

