Set Implicit Arguments.
Require Import MyTactics.
Require Import MyArithmetic.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import Abandon.
Require Import Row.
Require Import Intro.
Require Import Dup.
Require Import Typing.
Require Import JBasic.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are monotonic with respect to [rio]: actions by other
   threads cannot invalidate the permissions held by the current thread. *)

(* This lemma is presented as "Stability" in the paper. *)

Lemma j_rio:
(
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  consistent R1 ->
  forall R2,
  rio R1 R2 ->
  j f n R2 K P t T
) /\ (
  forall f n R1 K P tag Tss ts T,
  jbs f n R1 K P tag Tss ts T ->
  consistent R1 ->
  forall R2,
  rio R1 R2 ->
  jbs f n R2 K P tag Tss ts T
) /\ (
  forall R1 K P,
  jpi R1 K P ->
  consistent R1 ->
  forall R2,
  rio R1 R2 ->
  jpi R2 K P
).
Ltac rio_star_clear :=
  rio_star;
  repeat match goal with h: rio (star _ _) (star _ _) |- _ => clear h end.
Proof.
  apply j_jbs_jpi_mutind; intros; subst;
  (* Clear the original type derivation, which can mislead [eauto]
     when attempting to construct a new type derivation. *)
  repeat match goal with
  | h: j _ _ _ _ _ _ _ |- _ => clear h
  | h: jpi _ _ _ |- _ => clear h
  | h: is_locked_permission _ _ |- _ => inversion h; subst; revert h
         end; intros;
  try (rio_star_clear; try (rio_abandon_all; rio_star_clear));
  try econstructor (solve [
    eauto using @consistent_core, @rio_core, @SCAM_limit_rio,
                consistent_abandon_all, @SCAM_maps_stable, stable_mo2dx,
                stable_p2l, stable_p2hl,
                heap_project_morphism, lck_project_morphism
           with consistent_split lt_lt
  ]).
Qed.

(* [jcp] is stable. *)

Lemma jcp_rio:
  forall R1 t,
  jcp R1 t ->
  consistent R1 ->
  forall R2,
  rio R1 R2 ->
  jcp R2 t.
Proof.
  induction 1.
  (* Case [TThread]. *)
  eauto using (get1 j_rio) with j jcp.
  (* Case [TPar]. *)
  intros; subst; rio_star.
  eauto 6 with jcp consistent_split.
Qed.
  
