Set Implicit Arguments.
Require Import MyList.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.
Require Import Row.
Require Import Resources.
Require Import Signature.
Require Import Typing.

(* ---------------------------------------------------------------------------- *)

(* Inverting a [jbs] judgement. *)

(* This lemma represents a purely administrative induction over the list of
   branches, and is of no intrinsic interest. *)

Lemma jbs_inversion:
  (* If the branches of the [TSwitch] construct, beginning at the
     branch numbered [base], are well-typed, *)
  forall f n R K P base Tss ts T,
  jbs f n R K P base Tss ts T ->
  (* And if the tag [tag] that is of interest of us corresponds to
     the [diff]-th element in the list of branches [Tss], *)
  forall diff Ts,
  is_nth diff Tss Ts ->
  forall tag,
  base + diff = tag ->
  (* Then the list of branches [ts] can be put in the form [ctx[diff:=t]],
     where the branch [t] is well-typed. *)
  exists t ctx,
  row_context diff ctx /\
  fill_row ctx t = ts /\
  j f n R K (P tag Ts) t T.
Proof.
  introv h; dependent induction h; inversion 1; intros; subst.
  (* IsNthZero *)
  rewrite plus_0_r.
  xplit.
    econstructor; eauto.
    simpl_subst_goal. f_equal. eauto with subst_lift.
    eauto.
  (* IsNthSucc *)
  forwards: IHh.
    eauto.
    rewrite plus_Sn_m. rewrite plus_n_Sm. reflexivity.
  unpack; subst.
  xplit.
    econstructor; eauto.
    simpl_subst_goal. f_equal. eauto with subst_lift.
    eauto.
Qed.

Ltac jbs_inversion :=
  forwards: jbs_inversion; [ eauto | eauto | eapply plus_0_l | unpack; subst ];
  try row_context_equality_inversion.

