Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Resources.
Require Import WkInversion.
Require Import Exclusive.
Require Import Adopts.
Require Import Typing.
Require Import Top.

(* ------------------------------------------------------------------------- *)

(* [jctop] is defined as an abbreviation for [jpi _ _ (TyAnchored _ _)],
   which itself is essentially a conjunction of three judgements, [jc],
   [jpi], and [wk]. We need high-level lemmas and tactics that work at
   the level of [jctop] without exposing the lower level. *)

(* ------------------------------------------------------------------------- *)

(* The following tactic helps prove a lemma whose statement begins with a
   [jctop] hypothesis. The hypothesis is destructed, which gives rise to [jc],
   [jpi], and [wk] hypotheses; then, an induction is performed on the [jc]
   hypothesis. *)

Ltac jctop_induction :=
  (* Introduce the [jctop] hypothesis and destruct it. *)
  introv h; dependent destruction h; subst;
  (* We now have three hypotheses. Re-introduce them in the goal, in an
     appropriate order, so that the [jc] hypothesis becomes the primary
     hypothesis and can be used in an induction. *)
  match goal with h: wk _ _ _ |- _ => revert h end;
  match goal with h: jpi ?R2 _ _ |- _ => generalize R2 h; clear R2 h end;
  match goal with h: jc _ _ _ _ _ _ |- _ => revert h end;
  (* Attack the induction. *)
  intro h; dependent induction h;
  intros; wk_inversion; adopts_inversion; subst.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps prove a lemma whose statement begins with a
   [jctop] hypothesis. The hypothesis is destructed, which gives rise to [jc],
   [jpi], and [wk] hypotheses; only the first and last one are kept; then, an
   induction is performed on the [jc] hypothesis. *)

Ltac jctop_simplified_induction :=
  (* Introduce the [jctop] hypothesis and destruct it. *)
  introv h; dependent destruction h; subst;
  (* We now have three hypotheses. Drop the [jpi] hypothesis. Re-introduce
     the other two in the goal, in an appropriate order, so that the [jc]
     hypothesis becomes the primary hypothesis and can be used in an induction. *)
  match goal with h: wk _ _ _ |- _ => revert h end;
  match goal with h: jpi ?R2 _ _ |- _ => clear R2 h end;
  match goal with h: jc _ _ _ _ _ _ |- _ => revert h end;
  (* Attack the induction. *)
  intro h; dependent induction h;
  intros; wk_inversion; subst.

