Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.
Require Import Typing.
Require Import Resources.
Require Import WkInversion.
Require Import JBasic.
Require Import JCanonicalization.
Require Import JClassification.
Require Import JDecomposition.
Require Import Dup.
Require Import Intro.
Require Import SubsumptionSoundness.
Require Import DerivedSubtyping.
Require Import Top.
Require Import Harmless.
Require Import HarmlessDistrib.
Require Import DerivedTyping.

(* ---------------------------------------------------------------------------- *)

(* Decomposition of a unrestricted typing judgment for an application. *)

Lemma typed_app_decomposition:
  forall n Rh K P v t T,
  jt n Rh K P (TApp v t) T ->
  forall R Rp,
  wk K v VALUE ->
  jpi Rp K P ->
  star Rh Rp = R ->
  consistent R ->
  exists Rt Rv Pt Tt,
  jt n Rt K Pt t Tt /\
  jpi Rv K (TyStar (TyAnchored v (TyArrow Tt T)) Pt) /\
  star Rv Rt = R.
Proof.
(* TODO: still a few names variables in case [Frame]. *)

  (* Induction on the typing judgement. *)
  introv hjt.
  dependent induction hjt; intros; subst.
  (* Impossible case. *)
    wk_inversion.
  (* Frame. *)
    (* Split the resource [Rp]. *)
    jpi_inversion.
    (* Apply the induction hypothesis and split the resource [Rv]. *)
    forwards dec: IHhjt;
      eauto using @consistent_split_left, @consistent_by_equality, @star_associative.
    named_unpack dec.
    jpi_inversion.
    (* Split the goal. *)
    do 4 eexists; xplit.
    (* Typing of [t]. *)
      eauto with j.
    (* Justification of the permission. *)
      eapply JpiStar.
      (* Typing [v]. *)
        (* Canonicalize and decompose the typing judgement on [v]. *)
        make_jt_hypothesis hjv.
        forwards: canonicalization_preliminary hjv;
          eauto using @consistent_split_left, @consistent_by_equality, @consistent_1p23_12.
        classification.
        forwards: decomposition_TyArrow;
          eauto using @consistent_split_left, @consistent_by_equality, @consistent_1p23_12.
        unpack.
        (* Build the proof. *)
        econstructor.
        eauto with wk.
        apply JFunction;
          try unfold intro; simpl_lift_all; eauto with j.
        eauto with wk.
        (* instantiate (1 := Q); eauto with wk. *)
        (*   (* assert (wk K (TyMixStar x2 Q) TYPE); [ eauto with wk | eauto ]. *) *)
        eapply DupEmpty.
        eapply JpiEmpty.
        eauto using @core_right_unit.
      (* Resource and permissions conditions. *)
        eauto with j.
        auto.
    (* Resource equality. *)  
      repeat rewrite star_associative.
      rewrite <- dec.
      auto with aac.

  (* Cut. *)
    (* Apply the induction hypothesis. *)
    forwards dec: IHhjt;
      eauto using @consistent_by_equality with j aac.
    named_unpack dec.
    (* Build the proof. *)
    do 4 eexists; xplit.
    eauto using (get1 j_index_monotonic).
    eauto.
    rewrite dec. auto with aac.

  (* SubLeft *)
    (* Apply the induction hypothesis. *)
    forwards: IHhjt;
      eauto using jps_sound with consistent_split.
    unpack.
    (* Build the proof. *)
    do 4 eexists; xplit;
      eauto using (get1 j_index_monotonic).

  (* SubRight. *)
    (* Apply the induction hypothesis. *)
    forwards dec: IHhjt; eauto.
    named_unpack dec.
    jpi_inversion.
    (* Build the proof. *)
    do 4 eexists; xplit.
    eauto using (get1 j_index_monotonic).
      (* The interesting case: justifying the permissions. *)
      eapply JpiStar.
      (* Preparation. *)
      make_jt_hypothesis hjv.
      forwards: canonicalization_preliminary hjv;
        eauto using @consistent_split_left, @consistent_by_equality.
      sub_inversion.
      (* Typing [v]. *)
      eapply sub_sound.
        eapply SubCoArrow; eauto with j wk.
        eauto.
        eauto using @consistent_split_left, @consistent_by_equality.
      (* Resources and permissions. *)
      eauto.
      eauto.
    auto.

  (* Application. *)
  do 4 eexists; xplit; eauto using (get1 j_index_monotonic) with aac.

  (* Fold. *)
  forwards dec: IHhjt; eauto.
  named_unpack dec.
  jpi_inversion.
  do 4 eexists; xplit.
    eauto using (get1 j_index_monotonic).
    eapply JpiStar.
    make_jt_hypothesis hjv.
    forwards: canonicalization_preliminary hjv;
      eauto using @consistent_split_left, @consistent_by_equality.
    eapply sub_sound; eauto using @consistent_split_left, @consistent_by_equality.
      eapply SubCoArrow.
      eauto with j.
      econstructor; simpl_lift_goal; eauto using JpsFold with j_weakening.
      eauto with wk.
    eauto.
    eauto.
    eauto.

  (* ExistsElim. *)
  (* This case is not meaningful anymore. *)

  (* ForallIntro. *)
    (* Destruct the hypothesis on a universal quantification. *)
    jpi_inversion.
    (* Apply the induction hypothesis. *)
    forwards dec: IHhjt; simpl_lift_all; eauto with j_substitution.
    named_unpack dec.
    rewrite <- dec in *.
    jpi_inversion.
    (* Destruct well_kindness and harmlessness hypothesis. *)
    inversion H0; subst.
    inversion H; subst. wk_inversion.
    (* Build the proof. *)
    do 4 eexists; xplit; eauto with j.
    econstructor; eauto with j.
      (* Build a [jctop] hypothesis. *)
      make_jt_hypothesis hjv.
      forwards hjvc: canonicalization_preliminary hjv; eauto with consistent_split j.
      forwards hj: JcTopForallIntro hjvc; eauto with wk.
      (* Conclusion with [distrib]. *)
      eauto using harmless_distrib with consistent_split.

Grab Existential Variables.
auto. auto. auto. auto. auto.

Qed.

(* Decomposition of a unrestricted typing judgement for a spawn term. *)

Lemma typed_spawn_decomposition:
  forall n Rh K P v1 v2 T,
  jt n Rh K P (TSpawn v1 v2) T ->
  wk K v1 VALUE ->
  wk K v2 VALUE ->
  forall R Rp,
  jpi Rp K P ->
  star Rh Rp = R ->
  consistent R ->
  exists Rv Tv1 Tv2 Ru Rpu Pu nu,
  jpi Rv K (TyStar (TyAnchored v1 (TyArrow Tv2 Tv1)) (TyAnchored v2 Tv2)) /\
  jpi Rpu K Pu /\
  jt nu Ru K Pu TUnit T /\
  star Rv (star Ru Rpu) = R.
Proof.
  introv hjt.
  dependent induction hjt; intros; subst.

  (* Impossible. *)
  wk_inversion.

  (* Frame. *)
  jpi_inversion.
  forwards dec: IHhjt;
    eauto using @consistent_split_left, @consistent_by_equality, @star_associative.
  named_unpack dec.
  do 7 eexists; xplit.
  eauto with j.
  eauto.
  eauto with j.
  rewrite reconfigure10; rewrite dec; aac_reflexivity.

  (* Cut. *)
  eauto using @consistent_by_equality with j aac.

  (* SubLeft. *)
  eauto using jps_sound with consistent_split.

  (* SubRight. *)
  forwards dec: IHhjt; eauto.
  named_unpack dec.
  repeat eexists; try rewrite <- dec; eauto with j.

  (* Fold. *)
  forwards dec: IHhjt; eauto.
  named_unpack dec.
  repeat eexists; try rewrite <- dec; eauto with j.

  (* Spawn. *)
  jpi_inversion.
  make_jt_hypothesis ht1.
  make_jt_hypothesis ht2.
  forwards: canonicalization_preliminary ht1;
    eauto with consistent_split.
  forwards: canonicalization_preliminary ht2;
    eauto with consistent_split.
  do 7 eexists; xplit.
  eauto with j.
  eapply JpiEmpty.
  eapply JUnit.
  rewrite star_commutative; rewrite core_right_unit; eauto.


  (* ExistsElim. *)
  (* This case is not meaningful anymore. *)
  
  (* ForallIntro. *)
  false; eauto using harmful_spawn.


Grab Existential Variables.
auto. auto. auto. auto. auto.

Qed.

(* Decomposition of a unrestricted typing judgement on a spawn term in
   a deep evaluation context. *)

Lemma spawn_in_context_decomposition:
  forall n Rh K P D v1 v2 T,
  deep_eval_context D ->
  wk K v1 VALUE ->
  wk K v2 VALUE ->
  jt n Rh K P (fill D (TSpawn v1 v2)) T ->
  wk K P PERM ->
  forall R Rp,
  jpi Rp K P ->
  star Rh Rp = R ->
  consistent R ->
  exists Rv Tv1 Tv2 nd Rd Rpd Pd,
  jpi Rv K (TyStar (TyAnchored v1 (TyArrow Tv2 Tv1)) (TyAnchored v2 Tv2)) /\
  jt nd Rd K Pd (fill D TUnit) T /\
  jpi Rpd K Pd /\
  star (star Rd Rpd) Rv = R.
Proof.
  (* Induction on the evaluation context. *)
  introv hd. 
  revert n Rh P T.
  induction hd.
  (* Hole context. *)
    intros; simpl_subst_all.
    forwards: typed_spawn_decomposition; eauto.
    unpack.
    repeat eexists; try rewrite star_commutative; eauto.
  (* Application context. *)
    intros; open_eval_context.
    (* Decomposition of the application. *)
    forwards dec1: typed_app_decomposition; eauto using wk_empty.
    named_unpack dec1.
    jpi_inversion.
      (* Regroup the resources that concern the right member of the
         application. *)
      rewrite reconfigure09 in dec1.
    (* Apply the induction hypothesis. *)
    forwards dec2: IHhd;
      eauto using @consistent_split_right, @consistent_by_equality with wk.
    named_unpack dec2.
    jpi_inversion.
    (* Built the proof. *)
    repeat eexists.
    eauto with j.
    eauto with j.
    eauto with j.
    rewrite <- dec1; rewrite <- dec2. aac_reflexivity.
Qed.
