Set Implicit Arguments.
Require Import MyTactics.

(* ---------------------------------------------------------------------------- *)

(* Performing well-founded induction simultaneously over two statements. *)

(* Hoisting a universal quantifier out of a binary conjunction. *)

Lemma hoist_universal_2:
  forall (A : Type) (P Q : A -> Prop),
  (
    forall n, (P n /\ Q n)
  ) ->
  (forall n, P n) /\ (forall n, Q n).
Proof.
  introv h. split; intro.
  eapply get1. eauto.
  eapply last2. eauto.
Qed.

(* Pushing a universal quantifier into a conjunction (and projecting). *)

Lemma push_universal_2_1:
  forall (A : Type) (C P Q : A -> Prop),
  (forall y, C y -> P y /\ Q y) ->
   forall y, C y -> P y.
Proof.
  intros. eapply get1. eauto.
Qed.

Lemma push_universal_2_2:
  forall (A : Type) (C P Q : A -> Prop),
  (forall y, C y -> P y /\ Q y) ->
   forall y, C y -> Q y.
Proof.
  intros. eapply last2. eauto.
Qed.

(* Proving a conjunction of two universally quantified assertions
   by a simultaneous well-founded induction. *)

Lemma simultaneous_well_founded_induction_2:
  forall (P Q : nat -> Prop),
  (
    forall n,
    (forall m, m < n -> P m) ->
    (forall m, m < n -> Q m) ->
    P n
  ) ->
  (
    forall n,
    (forall m, m < n -> P m) ->
    (forall m, m < n -> Q m) ->
    Q n
  ) ->
  (
    forall n, P n
  ) /\ (
    forall n, Q n
  ).
Proof.
  introv h1 h2.
  eapply hoist_universal_2; intro n.
  induction_wf ih: lt_wf n.
  split; [ eapply h1 | eapply h2 ]; solve [
    eapply push_universal_2_1; eauto
  | eapply push_universal_2_2; eauto
  ].
Qed.

(* ---------------------------------------------------------------------------- *)

(* Same gig, this time over three statements. *)

Lemma hoist_universal_3:
  forall (A : Type) (P Q R : A -> Prop),
  (
    forall n, (P n /\ Q n /\ R n)
  ) ->
  (forall n, P n) /\ (forall n, Q n) /\ (forall n, R n).
Proof.
  introv h. xplit; intro.
  eapply get1. eauto.
  eapply get1. eapply last2. eauto.
  eapply last3. eauto.
Qed.

(* Pushing a universal quantifier into a conjunction (and projecting). *)

Lemma push_universal_3_1:
  forall (A : Type) (C P Q R : A -> Prop),
  (forall y, C y -> P y /\ Q y /\ R y) ->
   forall y, C y -> P y.
Proof.
  intros. eapply get1. eauto.
Qed.

Lemma push_universal_3_2:
  forall (A : Type) (C P Q R : A -> Prop),
  (forall y, C y -> P y /\ Q y /\ R y) ->
   forall y, C y -> Q y.
Proof.
  intros. eapply get2. eauto.
Qed.

Lemma push_universal_3_3:
  forall (A : Type) (C P Q R : A -> Prop),
  (forall y, C y -> P y /\ Q y /\ R y) ->
   forall y, C y -> R y.
Proof.
  intros. eapply last3. eauto.
Qed.

Lemma simultaneous_well_founded_induction_3:
  forall (P Q R : nat -> Prop),
  (
    forall n,
    (forall m, m < n -> P m) ->
    (forall m, m < n -> Q m) ->
    (forall m, m < n -> R m) ->
    P n
  ) ->
  (
    forall n,
    (forall m, m < n -> P m) ->
    (forall m, m < n -> Q m) ->
    (forall m, m < n -> R m) ->
    Q n
  ) ->
  (
    forall n,
    (forall m, m < n -> P m) ->
    (forall m, m < n -> Q m) ->
    (forall m, m < n -> R m) ->
    R n
  ) ->
  (
    forall n, P n
  ) /\ (
    forall n, Q n
  ) /\ (
    forall n, R n
  ).
Proof.
  introv h1 h2 h3.
  eapply hoist_universal_3; intro n.
  induction_wf ih: lt_wf n.
  xplit; [ eapply h1 | eapply h2 | eapply h3 ]; solve [
    eapply push_universal_3_1; eauto
  | eapply push_universal_3_2; eauto
  | eapply push_universal_3_3; eauto
  ].
Qed.

