Set Implicit Arguments.
Require Import ProofIrrelevance.

(* This is a variant of the lemmas [subset_eq_compat] and [subsetT_eq_compat]
   defined in the standard library module [ProofIrrelevanceFacts]. *)

Lemma subset_eq_compat:
  forall (U : Type) (P : U -> Prop) (x y : U) (p : P x) (q : P y),
  x = y ->
  exist P x p = exist P y q.
Proof.
  (* The proof is simpler than the one used in the standard library.
     Thanks to Xavier Leroy for suggesting the use of [subst]. *)
  intros. subst. f_equal. eapply proof_irrelevance.
Qed.

(* This tactic simplifies all hypotheses of the form [exist _ _ _ = exist _ _ _]. *)

Ltac exist_exist_hypotheses :=
  repeat match goal with h: exist _ ?x _ = exist _ ?y _ |- _ =>
    assert (x = y); [ congruence | clear h ]
  end.

(* This tactic simplifies a goal of the form [exist _ _ _ = exist _ _ _]. *)

Ltac exist_exist_goal :=
  try eapply subset_eq_compat.

