Require Import MyTactics.
Require Import Syntax.
Require Import Typing.
Require Import Resources.
Require Import Row.
Require Import Abandon.
Require Import Dup.
Require Import DeBruijn.

(* ----------------------------------------------------------------- *)

(* A set of notations and derived rules for references, for use in
   a simplified presentation of the formalization. *)

(* Reference. *)

Notation TRef :=
  (TLoc).

Lemma WkTRef:
  forall K l,
  wk K (TRef l) VALUE.
Proof.
  auto with wk.
Qed.

Notation ref2dx v :=
  (mtagvs2dx Exclusive 0 (RowCons v RowNil)).

(* Type of a reference. *)

Notation TyRef T :=
  (TyBlock Exclusive 0 (RowCons T RowNil) TyBottom).

Lemma WkTyRef:
  forall K T,
  wk K T TYPE ->
  wk K (TyRef T) TYPE.
Proof.
  auto with wk.
Qed.

(* Typing rule for references. *)

Lemma JRef:
  forall n f R R1 R2 P K l v T,
  (* The resource [R] is split in two parts. *)
  star R1 R2 = R ->
  (* One justifies that the closed value [v] has type [T]. *)
  wk zilch v VALUE ->
  jpi R1 K (TyAnchored v T) ->
  (* The other covers the memory block at address [l]. *)
  (* This block must not have adoptees. *)
  heap_maps R2 l (ref2dx v) ->
  no_adoptees R2 l ->
  (* Then we obtain the specialized reference type. *)
  j f n R K P (TRef l) (TyRef T).
Proof.
  introv hr hv hjpi hmaps hna.
  econstructor.
    eapply hr.
    eapply WkRowCons; [ eapply hv | eapply WkRowNil ].
    simpl. econstructor; [ eapply hjpi | eapply JpiEmpty | eapply @core_right_unit ].
    eapply AbandonAllZero. eapply hna.
    eapply @core_left_unit.
    simpl. eapply JpiEmpty.
    eauto with wk.
    eapply hmaps.
Qed.

(* Creation of a reference. *)

Notation TNewRef v :=
  (TNew 0 (RowCons v RowNil)).

Lemma WkTNewRef:
  forall K v,
  wk K v VALUE ->
  wk K (TNewRef v) PROG.
Proof.
  intros. eapply WkTNew. eauto with wk.
Qed.

Lemma JNewRef:
  forall n R K v T,
  wk K v VALUE ->
  wk K T TYPE ->
  jt (S n) R K (TyAnchored v T) (TNewRef v) (TyRef T).
Proof.
  intros.
  eapply (@JSubLeft 0).
  econstructor.
    eauto with wk.
    eauto with wk.
    simpl; auto.
    econstructor. econstructor.
    econstructor. econstructor.
    eauto with j.
Qed.

(* Reading a reference. *)

Notation TDeref v :=
  (TRead v 0).

Lemma WkTDeref:
  forall K v,
  wk K v VALUE ->
  wk K (TDeref v) PROG.
Proof.
  auto with wk.
Qed.

Lemma JDeref:
  forall n R K P v T,
  wk K v VALUE ->
  P = TyAnchored v (TyRef T) ->
  dup T ->
  jt n R K P (TDeref v) (TyMixStar T P).
Proof.
  intros.
  econstructor. 
    eauto.
    eauto.
    instantiate (1 := RowCons (TVar 0) RowNil).
      simpl_subst_goal. auto.
    econstructor; instantiate (1 := RowNil).
      simpl_lift_goal. auto.
    auto.
Qed.

(* Writing in a reference. *)

Notation TAssign v1 v2 :=
  (TWrite v1 0 v2).

Lemma WkTAssign:
  forall K v1 v2,
  wk K v1 VALUE ->
  wk K v2 VALUE ->
  wk K (TAssign v1 v2) PROG.
Proof.
  auto with wk.
Qed.

Lemma JAssign:
  forall n R K v1 v2 T1 T2,
  wk K v1 VALUE ->
  wk K v2 VALUE ->
  jt n R K
     (TyStar (TyAnchored v1 (TyRef T1)) (TyAnchored v2 T2))
     (TAssign v1 v2)
     (TyMixStar TyUnit (TyAnchored v1 (TyRef T2))).
Proof.
  econstructor.
    instantiate (1 := RowCons (TVar 0) RowNil).
      simpl_subst_goal. auto.
      simpl_subst_goal. auto.
    econstructor; instantiate (1 := RowNil).
      simpl_lift_goal. auto.
    auto.
    auto.
Qed.
