Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import MSASAMOperations.
Require Import Syntax.
Require Import Row.

(* ------------------------------------------------------------------------- *)

(* The following tactic helps destruct the well-kindedness judgement. *)

Ltac select_wk action :=
  (* We apply a number of lemmas at this particular point, inside
     the loop executed by [gji], because the use of these lemmas and
     the destruction of [wk] facts are mutually beneficial: each of
     them enables, in some situations, the other. *)
  match goal with
    (* Apply the lemma [wk_weakening_inverse] if possible. *)
  | h: wk (insert ?x _ _) (shift ?x _) _ |- _ =>
      forwards: wk_weakening_inverse; [ eexact h | reflexivity | clear h ]
    (* Apply the lemmas [row_context_wk_inversion_*] if possible. *)
  | h1: row_context _ ?Ts,
    h2: wk _ (subst _ 0 ?Ts) (ROW _) |- _ =>
      forwards: wk_row_context_inverse_hole; [ eexact h1 | eexact h2 |
      forwards: wk_row_context_inverse_rest; [ eexact h1 | eexact h2 |
      clear h1; clear h2
    ]]
    (* Apply the lemma [everywhere_maps] if possible. *)
  | h1: SAM_everywhere ?proj wkb ?h,
    h2: SAM_maps ?proj ?h _ _ |- _ =>
      forwards: SAM_everywhere_maps; [eexact h1 | eexact h2 | revert h2 ]
  (*   (* Apply the lemme [wk_eval_context_inversion] if possible. *) *)
  (* | h1: eval_context ?E, *)
  (*   h2: wk zilch (fill ?E _) _ |- _ => *)
  (*     forwards: wk_eval_context_inversion; [ eexact h1 | eexact h2 | *)
  (*     clear h1; clear h2; unpack *)
  (*   ] *)
  (*   (* Apply the lemme [wk_deep_eval_context_inversion] if possible. *) *)
  (* | h1: deep_eval_context ?D, *)
  (*   h2: wk zilch (fill ?D _) _ |- _ => *)
  (*     forwards: wk_deep_eval_context_inversion; [ eexact h1 | eexact h2 | *)
  (*     clear h1; clear h2; unpack *)
  (*   ] *)
    (* Destruct a well-kindedness judgement if possible. *)
  | h: wk _ _ _ |- _ =>
      action h
  | h: wkb _ |- _ =>
      action h
  | h: wkc _ |- _ =>
      action h
  | h: wkcp _ |- _ =>
      action h
  | h: wkh _ |- _ =>
      action h
  end.

Ltac wk_inversion :=
  gji select_wk.

(* ------------------------------------------------------------------------- *)

(* Sometimes, inverting a hypothesis of the form [wk _ t _] produces two
   sub-cases, the first of which corresponds to [WkValuesAreTerms] and
   can be immediately eliminated because [t] is not a value. *)

Ltac wkii :=
  match goal with h: wk _ ?t _ |- _ =>
    inversion h; subst; [
      solve [ match goal with h: wk _ t VALUE |- _ => inversion h end ]
    | clear h ]
  end.

(* ------------------------------------------------------------------------- *)

(* The tactic [wk_inversion] is a forward tactic. We sometimes need to
   reason about well-kindedness in backwards mode, as part of an [eauto]
   proof search. For this purpose, see [wk_inversion] in [Syntax.v]. *)

