Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import WkInversion.

(* A type or permission (of kind [TYPE] or [PERM]) admits at most one kind. *)

(* This property is currently not exploited, but I am keeping it around as a
   sanity check. *)

(* ------------------------------------------------------------------------- *)

(* Unfortunately, it is not the case that every term/type admits at most one
   kind, because [VALUE] is a sub-kind of [PROG]. The following predicate
   identifies the kinds that admit a uniqueness property. *)

Inductive unique : kind -> Prop :=
| UniqueType:
    unique TYPE
| UniquePerm:
    unique PERM.

Hint Constructors unique : wk.

Ltac select_unique action :=
  match goal with h: unique _ |- _ => action h end.
Ltac unique_inversion :=
  gji select_unique.

(* ------------------------------------------------------------------------- *)

(* The following lemma claims that [VALUE] does not satisfy [unique]. It is
   stated in a manner that allows it to be used by [eauto] below. *)

Lemma not_unique_VALUE:
  forall kappa1 kappa2,
  kappa1 = VALUE ->
  unique kappa1 ->
  kappa1 = kappa2.
Proof.
  intros. subst. unique_inversion.
Qed.

(* ------------------------------------------------------------------------- *)

(* The main lemma claims, if [kappa] satisfies [unique], then a type of
   kind [kappa] does not admit any other kind. *)

Lemma wk_unique:
  forall K T kappa1,
  wk K T kappa1 ->
  forall kappa2,
  wk K T kappa2 ->
  unique kappa1 ->
  kappa1 = kappa2.
Proof.
  induction 1; introv hwk ?; unique_inversion;
  dependent induction hwk; solve [
    congruence
  | wk_inversion
  | eauto using not_unique_VALUE with wk
  ].
Qed.

(* ------------------------------------------------------------------------- *)

(* This tactic helps exploit the lemma. *)

Ltac wk_unique :=
  match goal with
  h1: wk ?K ?T ?kappa1, h2: wk ?K ?T ?kappa2 |- _ =>
    first [
      forwards: (wk_unique h1 h2); [ econstructor | subst; clear h2 ]
    | forwards: (wk_unique h2 h1); [ econstructor | subst; clear h2 ]
    ]
  end.

